@@ -4,7 +4,7 @@ import (
 	"context"
 	"errors"
 	"fmt"
-	"net/netip"
+	"net"
 	"strings"
 	"time"
 
@@ -116,18 +116,12 @@ func NewSupervisor(config *TunnelConfig, orchestrator *orchestration.Orchestrato
 		gracefulShutdownC: gracefulShutdownC,
 		connAwareLogger:   log,
 	}
-	if useDatagramV2(config) {
-		// TODO: TUN-6654 listen for IPv6 and decide if it should listen on specific IP
-		listenIP, err := netip.ParseAddr("0.0.0.0")
-		if err != nil {
-			return nil, err
-		}
-		icmpProxy, err := ingress.NewICMPProxy(listenIP, config.Log)
-		if err != nil {
-			log.Logger().Warn().Err(err).Msg("Failed to create icmp proxy, will continue to use datagram v1")
-		} else {
-			edgeTunnelServer.icmpProxy = icmpProxy
-		}
+
+	icmpRouter, err := ingress.NewICMPRouter(config.Log)
+	if err != nil {
+		log.Logger().Warn().Err(err).Msg("Failed to create icmp router, ICMP proxy feature is disabled")
+	} else {
+		edgeTunnelServer.icmpRouter = icmpRouter
 	}
 
 	useReconnectToken := false
@@ -157,10 +151,14 @@ func (s *Supervisor) Run(
 	ctx context.Context,
 	connectedSignal *signal.Signal,
 ) error {
-	if s.edgeTunnelServer.icmpProxy != nil {
+	if s.edgeTunnelServer.icmpRouter != nil {
 		go func() {
-			if err := s.edgeTunnelServer.icmpProxy.Serve(ctx); err != nil {
-				s.log.Logger().Err(err).Msg("icmp proxy terminated")
+			if err := s.edgeTunnelServer.icmpRouter.Serve(ctx); err != nil {
+				if errors.Is(err, net.ErrClosed) {
+					s.log.Logger().Info().Err(err).Msg("icmp router terminated")
+				} else {
+					s.log.Logger().Err(err).Msg("icmp router terminated")
+				}
 			}
 		}()
 	}
@@ -435,15 +433,3 @@ func (s *Supervisor) authenticate(ctx context.Context, numPreviousAttempts int)
 	registrationOptions.NumPreviousAttempts = uint8(numPreviousAttempts)
 	return rpcClient.Authenticate(ctx, s.config.ClassicTunnel, registrationOptions)
 }
-
-func useDatagramV2(config *TunnelConfig) bool {
-	if config.NamedTunnel == nil {
-		return false
-	}
-	for _, feature := range config.NamedTunnel.Client.Features {
-		if feature == FeatureDatagramV2 {
-			return true
-		}
-	}
-	return false
-}
